from django.utils import timezone
from django.conf import settings
from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import AbstractUser, UserManager
from django.db import models
from garpix_notify.mixins import UserNotifyMixin
from garpix_utils.managers import ActiveManager
from garpix_utils.models import DeleteMixin
from garpix_utils.string import get_random_string
from phonenumber_field.modelfields import PhoneNumberField

from garpix_user.mixins.models.confirm import UserEmailConfirmMixin, UserPhoneConfirmMixin
from django.db.utils import IntegrityError
from django.utils.translation import gettext as _
import string

from garpix_user.models.password_history import PasswordHistory
from garpix_user.utils.current_date import set_current_date


class GarpixUser(DeleteMixin, UserEmailConfirmMixin, UserPhoneConfirmMixin, UserNotifyMixin, AbstractUser):
    is_email_confirmed = models.BooleanField(_("Email confirmed"), default=True)

    phone = PhoneNumberField(_("Phone number"), blank=True, default='')
    is_phone_confirmed = models.BooleanField(_("Phone number confirmed"), default=True)
    is_blocked = models.BooleanField(_("User is blocked"), default=False)
    login_attempts_count = models.IntegerField(_("Invalid log in attempts"), default=0)
    password_updated_date = models.DateTimeField(_("Password last updated date"), default=timezone.now)
    needs_password_update = models.BooleanField(_("Needs password update"), default=False)

    USERNAME_FIELDS = ('username',)

    objects = UserManager()
    active_objects = ActiveManager()

    class Meta:
        verbose_name = _('User')
        verbose_name_plural = _('Users')
        abstract = True

    def save(self, *args, **kwargs):
        is_new = self.pk is None
        GARPIX_USER_SETTINGS = getattr(settings, "GARPIX_USER", {})

        if is_new and not self.is_email_confirmed and GARPIX_USER_SETTINGS.get('USE_EMAIL_CONFIRMATION', False) and GARPIX_USER_SETTINGS.get(
                'USE_EMAIL_LINK_CONFIRMATION', True):

            confirmation_code = get_random_string(self.get_confirm_code_length('email'), string.digits)

            self.email_confirmation_code = confirmation_code
            self.email_code_send_date = set_current_date()
            self.new_email = self.email
            self.send_email_confirmation_link()

        super().save(*args, **kwargs)

    def delete(self, using=None, keep_parents=False):
        self.is_deleted = True
        self.is_active = False
        self.save()

    def __str__(self):
        return f'{self.first_name}, {self.last_name}' if self.first_name or self.last_name else self.email or str(self.phone)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if len(self.USERNAME_FIELDS) == 0:
            raise IntegrityError(_('USERNAME_FIELDS can\'t be empty'))
        for field in self.USERNAME_FIELDS:
            if field not in ('email', 'phone', 'username'):
                raise IntegrityError(
                    _(f'{field} can\'t be used as USERNAME_FIELDS. Only ("email", "phone", "username") supported'))

    def set_user_session(self, request):
        from garpix_user.models import UserSession
        current_user_session = UserSession.get_from_request(request)
        if current_user_session:
            if (user_user_session := UserSession.objects.filter(
                    user=self).first()) and user_user_session != current_user_session:
                current_user_session.delete()
            else:
                current_user_session.user = self
                current_user_session.recognized = UserSession.UserState.REGISTERED
                current_user_session.save()

    def set_password(self, raw_password):
        super().set_password(raw_password)
        if self.username:
            self.password_updated_date = set_current_date()
            self.save()
            password_history = PasswordHistory(user=self)
            password_history.password = make_password(raw_password)
            password_history.save()



from django.contrib.auth import get_user_model
from django.db.models import Q

from garpix_user.utils.get_password_settings import get_password_settings


class CustomAuthenticationBackend:
    def authenticate(self, request, username=None, password=None):
        password_settings = get_password_settings()

        if username is None or password is None:
            return None
        try:
            query = Q()
            for field in get_user_model().USERNAME_FIELDS:
                query |= Q(**{field: username.lower()})
            user = get_user_model().active_objects.get(query)
            pwd_valid = user.check_password(password)
            if pwd_valid:
                user.login_attempts_count = 0
                user.save()
                return user
            user.login_attempts_count += 1
            if password_settings['available_attempt'] != -1 and user.login_attempts_count >= password_settings[
                    'available_attempt']:
                user.is_blocked = True
            user.save()
            return None
        except get_user_model().DoesNotExist:
            return None

    def get_user(self, user_id):
        try:
            return get_user_model().active_objects.get(pk=user_id)
        except get_user_model().DoesNotExist:
            return None



from django.conf import settings
from django.contrib.auth import get_user_model
from django.utils.module_loading import import_string
from drf_spectacular.utils import extend_schema
from rest_framework.generics import CreateAPIView

from garpix_user.utils.drf_spectacular import user_session_token_header_parameter

User = get_user_model()
RegistrationSerializer = import_string(settings.GARPIX_USER.get('REGISTRATION_SERIALIZER', 'garpix_user.serializers.RegistrationSerializer'))


@extend_schema(
    parameters=[
        user_session_token_header_parameter()
    ]
)
class RegistrationView(CreateAPIView):
    queryset = User.objects.all()
    serializer_class = RegistrationSerializer


registration_view = RegistrationView.as_view()



from garpix_user.mixins.serializers import ToLowerMixin


from rest_framework import serializers


class EmailPreConfirmSendSerializer(ToLowerMixin, serializers.Serializer):
    email = serializers.EmailField()


class EmailConfirmSendSerializer(ToLowerMixin, serializers.Serializer):
    email = serializers.EmailField(required=False)


class EmailConfirmCheckCodeSerializer(serializers.Serializer):
    email_confirmation_code = serializers.CharField(max_length=255)



from django.contrib.auth.password_validation import CommonPasswordValidator, MinimumLengthValidator, \
    UserAttributeSimilarityValidator
from rest_framework import serializers

from django.utils.translation import ugettext as _

from garpix_user.utils.get_password_settings import get_password_settings
from garpix_user.utils.repluralize import rupluralize


class PasswordSerializerMixin:

    def _validate_password(self, value):

        password_settings = get_password_settings()

        UserAttributeSimilarityValidator().validate(value)
        MinimumLengthValidator(min_length=password_settings['min_length']).validate(value)
        CommonPasswordValidator().validate(value)

        # check for min digits number
        password_digits_num = sum(c.isdigit() for c in value)
        if password_digits_num < password_settings['min_digits']:
            raise serializers.ValidationError(
                _('Password must contain at least {min_digits} {digits}.').format(
                    min_digits=password_settings['min_digits'],
                    digits=rupluralize(password_settings['min_digits'],
                                       _('digit,digits')))
            )

        # check for min char number
        password_chars_num = sum(c.isalpha() for c in value)
        if password_chars_num < password_settings['min_chars']:
            raise serializers.ValidationError(
                _('Password must contain at least {min_chars} {chars}.').format(
                    min_chars=password_settings['min_chars'],
                    chars=rupluralize(password_settings['min_chars'],
                                      _('char,chars')))
            )

        # check for uppercase letter
        if sum(c.isupper() for c in value) < password_settings['min_uppercase']:
            raise serializers.ValidationError(
                _('Password must contain at least {min_uppercase} uppercase {letters}.').format(
                    min_uppercase=password_settings['min_uppercase'],
                    letters=rupluralize(password_settings['min_uppercase'], _('letter,letters')))
            )

        # check for special symbols
        if (len(value) - (password_digits_num + password_chars_num)) < password_settings['min_special_symbols']:
            raise serializers.ValidationError(
                _('Password must contain at least {min_special_symbols} special {symbols}.').format(
                    min_special_symbols=password_settings['min_special_symbols'],
                    symbols=rupluralize(password_settings['min_special_symbols'], _('symbol,symbols')))
            )



from rest_framework import serializers

from garpix_user.mixins.serializers import ToLowerMixin, PasswordSerializerMixin
from django.utils.translation import gettext_lazy as _
from rest_framework.authentication import authenticate


class RestorePasswordSerializer(ToLowerMixin, serializers.Serializer):
    username = serializers.CharField(required=True, help_text=_('Email or phone number'))


class RestoreSetPasswordSerializer(ToLowerMixin, PasswordSerializerMixin, serializers.Serializer):
    new_password = serializers.CharField(max_length=255, required=True)
    username = serializers.CharField(required=True, help_text=_('Email or phone number'))
    restore_password_confirm_code = serializers.CharField(max_length=15, required=True)

    def validate_new_password(self, value):
        self._validate_password(value)
        return value


class RestoreCheckCodeSerializer(ToLowerMixin, serializers.Serializer):
    restore_password_confirm_code = serializers.CharField(max_length=15, required=True)
    username = serializers.CharField(required=True, help_text=_('Email or phone number'))


class ChangePasswordSerializer(PasswordSerializerMixin, serializers.Serializer):
    new_password = serializers.CharField(max_length=255, required=True)
    password = serializers.CharField(required=True)

    def validate_new_password(self, value):
        self._validate_password(value)
        return value

    def validate_password(self, value):
        user = self.context['request'].user
        if not user.check_password(value):
            raise serializers.ValidationError(
                _('Password is incorrect')
            )
        return value


class ChangePasswordUnauthorizedSerializer(PasswordSerializerMixin, serializers.Serializer):
    username = serializers.CharField(label=_("Username"), required=True, write_only=True)
    password = serializers.CharField(
        label=_("Password"),
        style={'input_type': 'password'},
        trim_whitespace=False,
        required=True,
        write_only=True
    )
    new_password = serializers.CharField(write_only=True, required=True)
    new_password_2 = serializers.CharField(write_only=True, required=True)

    def validate_new_password(self, value):
        self._validate_password(value)
        return value

    def validate_new_password_2(self, value):

        if value != self.initial_data.get('new_password'):
            raise serializers.ValidationError(_("Passwords do not match"))

        return value

    def validate(self, attrs):
        username = attrs.get('username')
        password = attrs.get('password')
        username = username.lower()
        request = self.context.get('request')

        user = authenticate(request=request,
                            username=username, password=password)

        if not user:
            msg = _('Unable to log in with provided credentials.')
            raise serializers.ValidationError(msg, code='authorization')
        if user.is_blocked:
            msg = _(
                'Your account is blocked. Please contact your administrator')
            raise serializers.ValidationError(msg, code='authorization')

        super().validate(attrs)

        attrs['user'] = user

        return attrs



from django.utils import translation


def rupluralize(value, arg='character,characters'):

    lang = translation.get_language()

    args = arg.split(",")
    number = abs(int(value))
    a = number % 10
    b = number % 100

    if lang == 'ru':

        if (a == 1) and (b != 11):
            return args[0]
        elif (a >= 2) and (a <= 4) and ((b < 10) or (b >= 20)):
            return args[1]
        else:
            return args[2]

    else:
        if a == 1:
            return args[0]
        else:
            return args[1]



from rest_framework.serializers import ModelSerializer

from garpix_user.models.user_session import UserSession


class UserSessionSerializer(ModelSerializer):
    class Meta:
        model = UserSession
        fields = (
            '__all__'
        )


class UserSessionTokenSerializer(ModelSerializer):
    class Meta:
        model = UserSession
        fields = ('token_number',)



from solo.models import SingletonModel
from django.db import models
from django.utils.translation import gettext as _


class GarpixUserPasswordConfiguration(SingletonModel):
    min_length = models.PositiveIntegerField(default=12, verbose_name=_('Минимальная длина пароля'))
    min_digits = models.PositiveIntegerField(default=2, verbose_name=_('Минимальное количество цифр'))
    min_chars = models.PositiveIntegerField(default=2, verbose_name=_('Минимальное количество букв'))
    min_uppercase = models.PositiveIntegerField(default=1, verbose_name=_('Минимальное количество заглавных букв'))
    min_special_symbols = models.PositiveIntegerField(default=1, verbose_name=_('Минимальное количество спец.символов'))
    available_attempt = models.IntegerField(default=-1, help_text=_('-1 если ограничение не требуется'),
                                            verbose_name=_('Количество допустимых неуспешных попыток входа'))
    password_history = models.IntegerField(default=-1, help_text=_('-1 если ограничение не требуется'),
                                           verbose_name=_(
                                               'Количество последних паролей, которые нельзя использовать при смене'))
    password_validity_period = models.IntegerField(default=-1, help_text=_('-1 если ограничение не требуется'),
                                                   verbose_name=_(
                                                       'Срок действия пароля (в днях)')
                                                   )
    password_first_change = models.BooleanField(default=False,
                                                verbose_name=_('Обязательная смена пароля при первом входе'))
    password_validity_inform_days = models.IntegerField(default=-1,
                                                        help_text=_('-1 если отправка уведомлений не требуется'),
                                                        verbose_name=_(
                                                            'За сколько дней начинать отправлять уведомление об истечении срока пароля')
                                                        )
    access_token_ttl_seconds = models.IntegerField(default=0,
                                                   help_text=_('0 если ограничение не требуется'),
                                                   verbose_name=_(
                                                       'Время жизни авторизационного токена')
                                                   )
    refresh_token_ttl_seconds = models.IntegerField(default=0,
                                                    help_text=_('0 если ограничение не требуется'),
                                                    verbose_name=_(
                                                        'Время жизни рефреш токена')
                                                    )

    class Meta:
        verbose_name = 'Настройки безопасности входа'
        verbose_name_plural = 'Настройки безопасности входа'

    def __str__(self):
        return ''



from django.contrib import admin

from ..models.refferal import ReferralType


@admin.register(ReferralType)
class ReferralTypeAdmin(admin.ModelAdmin):
    list_display = ['title']
    fields = ['title']



from django.conf import settings
from django.db import models
from django.utils.translation import ugettext as _


class PasswordHistory(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, verbose_name=_('User'))
    password = models.CharField(_('password'), max_length=128)
    created_at = models.DateTimeField(auto_now_add=True, verbose_name=_('Date created'))

    class Meta:
        verbose_name = _('Password history')
        verbose_name_plural = _('Password history')



from django.conf import settings

from garpix_user.models import GarpixUserPasswordConfiguration


def get_password_settings():
    GARPIX_USER_SETTINGS = settings.GARPIX_USER

    _access_token_ttl_seconds = getattr(settings, 'GARPIX_ACCESS_TOKEN_TTL_SECONDS', 0)  # DEPRECATED
    _refresh_token_ttl_seconds = getattr(settings, 'GARPIX_REFRESH_TOKEN_TTL_SECONDS', 0)  # DEPRECATED

    if GARPIX_USER_SETTINGS.get('ADMIN_PASSWORD_SETTINGS', False):
        admin_password_settings = GarpixUserPasswordConfiguration.get_solo()
        min_length = admin_password_settings.min_length
        min_digits = admin_password_settings.min_digits
        min_chars = admin_password_settings.min_chars
        min_uppercase = admin_password_settings.min_uppercase
        min_special_symbols = admin_password_settings.min_special_symbols
        available_attempt = admin_password_settings.available_attempt
        password_history = admin_password_settings.password_history
        password_validity_period = admin_password_settings.password_validity_period
        password_first_change = admin_password_settings.password_first_change
        password_validity_inform_days = admin_password_settings.password_validity_inform_days
        access_token_ttl_seconds = admin_password_settings.access_token_ttl_seconds
        refresh_token_ttl_seconds = admin_password_settings.refresh_token_ttl_seconds
    else:
        min_length = GARPIX_USER_SETTINGS.get('MIN_LENGTH_PASSWORD', 8)
        min_digits = GARPIX_USER_SETTINGS.get('MIN_DIGITS_PASSWORD', 2)
        min_chars = GARPIX_USER_SETTINGS.get('MIN_CHARS_PASSWORD', 2)
        min_uppercase = GARPIX_USER_SETTINGS.get('MIN_UPPERCASE_PASSWORD', 1)
        min_special_symbols = GARPIX_USER_SETTINGS.get('MIN_SPECIAL_PASSWORD', 1)
        available_attempt = GARPIX_USER_SETTINGS.get('AVAILABLE_ATTEMPT', -1)
        password_history = GARPIX_USER_SETTINGS.get('PASSWORD_HISTORY', -1)
        password_validity_period = GARPIX_USER_SETTINGS.get('PASSWORD_VALIDITY_PERIOD', -1)
        password_first_change = GARPIX_USER_SETTINGS.get('PASSWORD_FIRST_CHANGE', False)
        password_validity_inform_days = GARPIX_USER_SETTINGS.get('PASSWORD_VALIDITY_INFORM_DAYS', -1)
        access_token_ttl_seconds = GARPIX_USER_SETTINGS.get('ACCESS_TOKEN_TTL_SECONDS', _access_token_ttl_seconds)
        refresh_token_ttl_seconds = GARPIX_USER_SETTINGS.get('REFRESH_TOKEN_TTL_SECONDS', _refresh_token_ttl_seconds)

    return {
        'min_length': min_length,
        'min_digits': min_digits,
        'min_chars': min_chars,
        'min_uppercase': min_uppercase,
        'min_special_symbols': min_special_symbols,
        'available_attempt': available_attempt,
        'password_history': password_history,
        'password_validity_period': password_validity_period,
        'password_first_change': password_first_change,
        'password_validity_inform_days': password_validity_inform_days,
        'access_token_ttl_seconds': access_token_ttl_seconds,
        'refresh_token_ttl_seconds': refresh_token_ttl_seconds
    }



from django.conf import settings

from .user_session import UserSessionAdmin  # noqa

from .user import UserAdmin  # noqa

from .password_history import PasswordHistoryAdmin  # noqa


if settings.GARPIX_USER.get('USE_REFERRAL_LINKS', False):
    from .referral_type import ReferralTypeAdmin  # noqa

if settings.GARPIX_USER.get('ADMIN_PASSWORD_SETTINGS', False):
    from .site_config import GarpixUserPasswordConfigurationAdmin  # noqa



from drf_spectacular.utils import extend_schema
from rest_framework import parsers, renderers

from garpix_user.mixins.views.auth_token_mixin import AuthTokenViewMixin
from garpix_user.serializers.auth_token_serializer import AuthTokenSerializer
from rest_framework.views import APIView
from django.conf import settings
from garpix_user.utils.drf_spectacular import user_session_token_header_parameter


@extend_schema(
    parameters=[
        user_session_token_header_parameter()
    ]
)
class ObtainAuthToken(AuthTokenViewMixin, APIView):
    throttle_classes = ()
    permission_classes = ()
    parser_classes = (parsers.FormParser, parsers.MultiPartParser, parsers.JSONParser,)
    renderer_classes = (renderers.JSONRenderer,)
    serializer_class = AuthTokenSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.serializer_class(data=request.data,
                                           context={'request': request})
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data['user']
        use_jwt = settings.GARPIX_USER.get('REST_AUTH_TOKEN_JWT', False)

        if use_jwt:
            return self._get_jwt_data(user)
        return self._get_access_token_data(user)


obtain_auth_token = ObtainAuthToken.as_view()



from rest_framework import parsers, renderers
from garpix_user.models.access_token import AccessToken as Token
from oauth2_provider.models import AccessToken
from rest_framework.response import Response
from rest_framework.views import APIView
from garpix_user.models.refresh_token import RefreshToken
from rest_framework.permissions import IsAuthenticated
from garpix_user.utils import get_token_from_request


class LogoutView(APIView):
    throttle_classes = ()
    permission_classes = (IsAuthenticated,)
    parser_classes = (parsers.FormParser, parsers.MultiPartParser, parsers.JSONParser,)
    renderer_classes = (renderers.JSONRenderer,)

    def post(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            token = get_token_from_request(request)
            if token is not None:
                Token.objects.filter(key=token).delete()
                AccessToken.objects.filter(token=token).delete()
                RefreshToken.objects.filter(key=token).delete()
                return Response({
                    'result': True,
                })
        return Response({
            'result': False,
        }, status=401)


logout_view = LogoutView.as_view()
